! Copyright 2019 Wojciech Kosior
  
! This is free and unencumbered software released into the public domain.

! Anyone is free to copy, modify, publish, use, compile, sell, or
! distribute this software, either in source code form or as a compiled
! binary, for any purpose, commercial or non-commercial, and by any
! means.

! In jurisdictions that recognize copyright laws, the author or authors
! of this software dedicate any and all copyright interest in the
! software to the public domain. We make this dedication for the benefit
! of the public at large and to the detriment of our heirs and
! successors. We intend this dedication to be an overt act of
! relinquishment in perpetuity of all present and future rights to this
! software under copyright law.

! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
! OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
! ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
! OTHER DEALINGS IN THE SOFTWARE.

! For more information, please refer to <http://unlicense.org/>

MODULE dotmat
  IMPLICIT none
  PRIVATE

  PUBLIC :: dotmull
  PRIVATE :: dotmull_4, dotmull_8, dotmull_16
  
  INTERFACE dotmull
     procedure dotmull_4, dotmull_8, dotmull_16
  END INTERFACE dotmull

CONTAINS

  FUNCTION dotmull_4(A, B) result(C)
    IMPLICIT none
    real(kind=4), intent(in), dimension(1:,1:) :: A, B
    real(kind=4), dimension(size(A, 1), size(B, 2)) :: C
    integer :: i, j

    DO j = 1, size(B, 2)
       DO i = 1, size(A, 1)

          C(i,j) = dot_product(A(i,:), B(:,j))
       END DO
    END DO
    
  END FUNCTION dotmull_4
  
  FUNCTION dotmull_8(A, B) result(C)
    IMPLICIT none
    real(kind=8), intent(in), dimension(1:,1:) :: A, B
    real(kind=8), dimension(size(A, 1), size(B, 2)) :: C
    integer :: i, j

    DO j = 1, size(B, 2)
       DO i = 1, size(A, 1)

          C(i,j) = dot_product(A(i,:), B(:,j))
       END DO
    END DO
    
  END FUNCTION dotmull_8
  
  FUNCTION dotmull_16(A, B) result(C)
    IMPLICIT none
    real(kind=16), intent(in), dimension(1:,1:) :: A, B
    real(kind=16), dimension(size(A, 1), size(B, 2)) :: C
    integer :: i, j

    DO j = 1, size(B, 2)
       DO i = 1, size(A, 1)

          C(i,j) = dot_product(A(i,:), B(:,j))
       END DO
    END DO
    
  END FUNCTION dotmull_16

END MODULE dotmat
